"""Module containing account related views."""

# vim: set tw=80 ts=4 sw=4 sts=4:

import dateutil.parser

from flask_restplus import Namespace, Resource, fields

from ..models import db
from ..models.accounts import Account
from ..models.operations import Operation

from .users import requires_auth


# pylint: disable=invalid-name
ns = Namespace('account', description='Account management')

# Account model.
account_model = ns.model('Account', {
    'id': fields.Integer(
        default=None,
        readonly=True,
        description='Id of the account'),
    'name': fields.String(
        required=True,
        description='Name of the account'),
    'authorized_overdraft': fields.Float(
        required=True,
        description='Authorized overdraft')
})

# Account status model.
solds_model = ns.model('Solds', {
    'current': fields.Float(
        readonly=True,
        description='Current sold of the account'),
    'pointed': fields.Float(
        readonly=True,
        description='Pointed sold of the account'),
    'future': fields.Float(
        readonly=True,
        description='Future sold of the account')
})

# Account balance model.
balance_model = ns.model('Balance', {
    'expenses': fields.Float(
        readonly=True,
        description='Total amount of expenses'),
    'revenues': fields.Float(
        readonly=True,
        description='Total amount of revenues'),
    'balance': fields.Float(
        readonly=True,
        description='Balance'),
})

# Category with expenses and revenues.
category_model = ns.model('Category', {
    'category': fields.String(
        readonly=True,
        description='Category name'),
    'expenses': fields.Float(
        readonly=True,
        description='Total amount of expenses for the category'),
    'revenues': fields.Float(
        readonly=True,
        description='Total amount of revenues for the category')
})

# OHLC model.
ohlc_model = ns.model('OHLC', {
    'operation_date': fields.DateTime(
        dt_format='iso8601',
        readonly=True,
        required=True,
        description='Date of the OHLC object'
    ),
    'open': fields.Float(
        readonly=True,
        required=True,
        description='Open value'
    ),
    'high': fields.Float(
        readonly=True,
        required=True,
        description='High value'
    ),
    'low': fields.Float(
        readonly=True,
        required=True,
        description='Low value'
    ),
    'close': fields.Float(
        readonly=True,
        required=True,
        description='Close value'
    )
})

# Parser for a date range.
range_parser = ns.parser()
range_parser.add_argument(
    'begin',
    type=lambda a: dateutil.parser.parse(a) if a else None,
    required=False,
    default=None,
    location='args',
    help='Begin date of the time period'
)
range_parser.add_argument(
    'end',
    type=lambda a: dateutil.parser.parse(a) if a else None,
    required=False,
    default=None,
    location='args',
    help='End date of the time period'
)


@ns.route('/')
@ns.doc(
    security='apikey',
    responses={
        401: 'Unauthorized'
    })
class AccountListResource(Resource):
    """Resource used to handle account lists."""

    @requires_auth
    @ns.response(200, 'OK', [account_model])
    @ns.marshal_list_with(account_model)
    def get(self):
        """ Returns accounts with their balances."""

        return Account.query().all(), 200

    @requires_auth
    @ns.expect(account_model)
    @ns.response(201, 'Account created', account_model)
    @ns.response(406, 'Invalid account data')
    @ns.marshal_with(account_model)
    def post(self):
        """Create a new account."""

        data = self.api.payload

        # A new account MUST NOT have an id;
        if 'id' in data and data['id']:
            ns.abort(
                406,
                error_message='Id must not be provided on creation.'
            )

        # Instantiate account with data.
        account = Account(**data)

        # Add new account in session.
        db.session.add(account)

        # Flush session to have id in account.
        db.session.flush()

        # Return account.
        return account, 201


@ns.route('/<int:id>')
@ns.doc(
    security='apikey',
    params={
        'id': 'Id of the account to manage'
    },
    responses={
        401: 'Unauthorized',
        404: 'Account not found'
    })
class AccountResource(Resource):
    """Resource to handle accounts."""

    @requires_auth
    @ns.response(200, 'OK', account_model)
    @ns.marshal_with(account_model)
    def get(self, id):
        """Get an account."""

        account = Account.query().get(id)

        if not account:
            ns.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        # Note: if we don't pass the code, the result is seen as a tuple and
        # causes error on marshalling.
        return account, 200

    @requires_auth
    @ns.expect(account_model)
    @ns.response(200, 'OK', account_model)
    @ns.response(406, 'Invalid account data')
    @ns.marshal_with(account_model)
    def post(self, id):
        """Update an account."""

        data = self.api.payload

        # Check ID consistency.
        if 'id' in data and data['id'] and data['id'] != id:
            ns.abort(
                406,
                error_message='Id must not be provided or changed on update.'
            )

        # Need to get the object to update it.
        account = Account.query().get(id)

        if not account:
            ns.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        # SQLAlchemy objects ignore __dict__.update() with merge.
        for k, v in data.items():
            setattr(account, k, v)

        db.session.merge(account)

        # Return account.
        return account, 200

    @requires_auth
    @ns.response(204, 'Account deleted', account_model)
    @ns.marshal_with(account_model)
    def delete(self, id):
        """Delete an account."""

        # Need to get the object to update it.
        account = Account.query().get(id)

        if not account:
            ns.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        db.session.delete(account)

        return None, 204


@ns.route('/<int:id>/solds')
class SoldsResource(Resource):
    """Resource to expose solds."""

    @requires_auth
    @ns.doc(
        security='apikey',
        responses={
            200: ('OK', solds_model),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @ns.marshal_with(solds_model)
    def get(self, id):
        """Get solds for a specific account and date range."""

        account = Account.query().get(id)

        if not account:
            ns.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        # Note: if we don't pass the code, the result is seen as a tuple and
        # causes error on marshalling.
        return account.solds(), 200


@ns.route('/<int:id>/balance')
class BalanceResource(Resource):
    """Resource to expose balances."""

    @requires_auth
    @ns.doc(
        security='apikey',
        responses={
            200: ('OK', balance_model),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @ns.expect(range_parser)
    @ns.marshal_with(balance_model)
    def get(self, id):
        """Get account balance for a specific date range."""

        account = Account.query().get(id)

        if not account:
            ns.abort(
                404,
                error_message='Account with id %d not found.' % id
            )

        data = range_parser.parse_args()

        # Note: if we don't pass the code, the result is seen as a tuple and
        # causes error on marshalling.
        return account.balance(**data), 200


@ns.route("/<int:id>/category")
class CategoryResource(Resource):
    """Resource to expose categories."""

    @requires_auth
    @ns.doc(
        security='apikey',
        responses={
            200: ('OK', [category_model]),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @ns.expect(range_parser)
    @ns.marshal_list_with(category_model)
    def get(self, id):
        """Get account category balances for a specific date range."""

        data = range_parser.parse_args()

        return Operation.get_categories_for_range(id, **data).all()


@ns.route('/<int:id>/ohlc')
class OHLCResource(Resource):
    """Resource to expose OHLC."""

    @requires_auth
    @ns.doc(
        security='apikey',
        responses={
            200: ('OK', [ohlc_model]),
            401: 'Unauthorized',
            404: 'Account not found'
        })
    @ns.expect(range_parser)
    @ns.marshal_list_with(ohlc_model)
    def get(self, id):
        """Get OHLC data for a specific date range and account."""

        data = range_parser.parse_args()

        return Operation.get_ohlc_per_day_for_range(id, **data).all()
